/***************************************************************************
 *   Copyright (C) 2004 by Joseph Wenninger                                *
 *   jowenn@kde.org                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>

#include <qdir.h>
#include <kprocess.h>
#include <kdebug.h>
#include "knxconnection.h"
#include "winuserpwd.h"
#include <assert.h>
#include <qfile.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "kcookie.h"
#include <klocale.h>
#include <qtimer.h>
#include <qdesktopwidget.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <kstandarddirs.h>
#include <kpassdlg.h>
#include <kmessagebox.h> //move the kmessagebox calles to the connecting dialog
#include <kinputdialog.h> // perhaps move to connecting dialog
#include <qstringlist.h>

KProcess *KNXConnection::SSHData::newProcess(QObject *parent) {
	if (host.isEmpty() || user.isEmpty() || port.isEmpty()) return 0;
		KProcess *proc=new KProcess(parent);
		(*proc)<<"ssh" << "-i" <<"/usr/share/knx/client.id_dsa.key" << (user+"@"+host)<<"-p"<<port;
		return proc;
	}

QString KNXConnection::SSHData::targetData() {return host+":"+port;}


KNXConnection::KNXConnection():QObject() {
	send0=0;
	send1=0;
	m_state=INITIALIZATION;
	m_data="";
	m_sshProcess=0;
	m_proxyProcess=0;
	m_ignoreAbortConnecting=false;
	m_connectionTimeout=0;
	m_tunneledConnection=0;
}

KNXConnection::~KNXConnection() {
	if (m_connecting) m_connecting->clearConnection();
	if (m_connectionTimeout) m_connectionTimeout->stop();
	if (m_sshProcess)
		disconnect(m_sshProcess,SIGNAL(processExited(KProcess*)),this,SLOT(childExited(KProcess*)));
}

bool KNXConnection::doConnect(const QString &sessionName) {
	kdDebug()<<"doConnect"<<endl;
	m_sessionName=sessionName;
	m_connecting=new ConnectingDialog(this,m_sessionName);
	m_connecting->setActivity(i18n("Trying to connect to <B>%1</B>").arg(m_sshData.targetData()));
	m_connecting->show();
	if (m_tunneledConnection)
		if (!m_tunneledConnection->askCredentials(m_connecting,m_sessionName)) {
		delete m_connecting;
		m_connecting=0;
		return false;
	}
#warning fixme, m_connecting cleanup

	QDesktopWidget desktop;
	long displayPlanes=DisplayPlanes(qt_xdisplay(),desktop.screenNumber(m_connecting));
	QRect screenRect=desktop.screenGeometry();
	m_screenInfo=QString("%1x%2x%3").arg(screenRect.width()).arg(screenRect.height()).arg(displayPlanes);
	KCookie cookieinfo;
	m_cookie=cookieinfo.displayAuth();
	if (m_cookie.find("MIT-MAGIC-COOKIE-1 ")!=0) {
		//error
	} else {
		m_cookie=m_cookie.right(m_cookie.length()-/*length MIT-MAGIC-....*/19);
	}

	kdDebug()<<"Trying to start ssh"<<endl;
	m_sshProcess=m_sshData.newProcess(this);
	if (!m_sshProcess) return false;
	//(*m_sshProcess)<<"ssh" << "-i" <<"/usr/share/knx/client.id_dsa.key" << m_sshData.connectionString());
	connect(m_sshProcess,SIGNAL(receivedStdout(KProcess *, char *, int)),
		this,SLOT(receivedFromSSH(KProcess*, char*,int)));
	connect(m_sshProcess,SIGNAL(wroteStdin(KProcess *)),this,SLOT(dataHasBeenWrittenToSSH()));
	connect(m_sshProcess,SIGNAL(processExited(KProcess*)),this,SLOT(childExited(KProcess*)));
	m_connectionTimeout=new QTimer(this);
	connect(m_connectionTimeout,SIGNAL(timeout()),this,SLOT(timeoutWhileConnecting()));
	m_connectionTimeout->start(20000); /*20 seconds ssh startup timeout */
	m_sshProcess->setUsePty(KProcess::All,true);
	m_sshProcess->start(KProcess::NotifyOnExit,(KProcess::Communication)(KProcess::Stdin | KProcess::Stdout));
	return true;
}


void KNXConnection::setSSHConnection(const QString& user, const QString &host, const QString &port) {
	m_sshData.user=user;
	if (!port.isEmpty()) m_sshData.host=host;
	if (!host.isEmpty()) m_sshData.port=port;
}

void KNXConnection::setNXLogin(const QString& password, const QString& username) {
	if (!username.isEmpty()) m_username=username;
	m_password=password;
}


void KNXConnection::receivedFromSSH (KProcess *proc, char *buffer, int buflen) {
	QCString newData(buffer,buflen+1);
	m_data+=newData;
	if (! m_data.data()) return;
	kdDebug()<<"receivedFromSSH: "<<newData<<endl;
	switch (m_state) {
		case INITIALIZATION:
					if (findLineStartingWith("HELLO NXSERVER")) {
						if (waitsForInput()) {
							if (m_connecting) m_connecting->setActivity(i18n("Setting initial connection parameters"));
							sendString("hello NXCLIENT - Version 1.4.0\n");
							m_state=PREAMBLE_SHELL;
							m_connectionTimeout->changeInterval(5000);
						}
					} else if (m_data.right(2) == "? ") {
						m_connectionTimeout->stop();
						if (KMessageBox::Yes==KMessageBox::warningYesNo(m_connecting,m_data,i18n("SSH-Server security"))) {
							sendString("yes\n");
							m_connectionTimeout->changeInterval(5000);
						} else {
							sendString("no\n");
							m_state=ERROR;
							if (m_connecting) m_connecting->badServerKey(m_sshData.targetData());
						}
						m_data="";	
					}
			break;
		case PREAMBLE_SHELL:
					if (gotResponse("522")) {
						//error
					} else if (gotResponse("134")) {
						if (waitsForInput()) {
							sendString("SET SHELL_MODE SHELL\n");
							m_state=PREAMBLE_PASSWORD;
							m_connectionTimeout->changeInterval(5000);
						}
					}
				//if (waitsForInput()
			break;
		case PREAMBLE_PASSWORD:
					if (waitsForInput()) {
							sendString("SET AUTH_MODE PASSWORD\n");
							m_state=LOGIN;
							m_connectionTimeout->changeInterval(5000);
						}
			break;
		case LOGIN:
					if (waitsForInput()) {
							if (m_connecting) m_connecting->setActivity(i18n("Authenticating user"));
							sendString("login\n");
							m_state=LOGIN_NAME;
							m_connectionTimeout->changeInterval(5000);
					}
			break;
		case LOGIN_NAME:
					if (waitsForInput(101)) {
							sendString(m_username+"\n");
							m_state=LOGIN_PASSWORD;
							m_connectionTimeout->changeInterval(5000);
						}
			break;
		case LOGIN_PASSWORD:
					if (waitsForInput(102)) {
							sendString(m_password+"\n");
							m_state=STARTSESSION;
							m_connectionTimeout->changeInterval(5000);
					} else if (gotResponse("500")) {
						m_connectionTimeout->stop();
						m_state=ERROR;
						if (m_connecting) m_connecting->accessDenied(true,m_sshData.targetData());
						deleteLater();
					}

			break;
		case STARTSESSION: 
					{
						if (gotResponse("103")) {
							if (waitsForInput()) {
								if (m_connecting) m_connecting->setActivity(i18n("Starting remote session"));
								sendString("startsession --session=\""+m_sessionName+"\" "
									+m_sessionString+" --screeninfo=\""+m_screenInfo+"+render"
									+QString("\" --cookie=\"%1\" ").arg(m_cookie)+
									(m_tunneledConnection ? m_tunneledConnection->settingsString():"")+"\n"

							/*	"startsession --session=\"SomeHost\" --type=\"unix-kde\" --cache=\"8M\" --images=\"32M\" --cookie=\""+
								m_cookie +"\" --link=\"modem\" --backingstore=\"when_requested\" --geometry=\"800x600+112+62\" --keyboard=\"us\""
								" --kbtype=\"pc105/us\" --media=\"0\" --agent_server=\"\" --agent_user=\"\" --agent_password=\"\" "
								" --screeninfo=\"800x600x24+render\"\n"*/
								);
								m_state=GETID;
								m_connectionTimeout->changeInterval(10000);
							}

						} else if (gotResponse("404")) {
							m_connectionTimeout->stop();
							m_state=ERROR;
							if (m_connecting) m_connecting->accessDenied(true,m_sshData.targetData());
							deleteLater();
						}
					}

			break;
		case GETID:
				if (getValue(700,m_sessionID) ) 
					m_state=GETDISPLAY; else break;
			
		case GETDISPLAY:
				if (getValue(705,m_sessionDisplay) ) 
					m_state=GETCOOKIE; else break;
					
			
		case GETCOOKIE:
				if (getValue(701,m_sessionCookie) ) m_state=SESSION; else break;
			
		case SESSION: {
					m_connectionTimeout->stop();
					if (m_connecting) m_connecting->setActivity(i18n("Opening NX Display"));
					kdDebug()<<"Creating session file"<<endl;
					KStandardDirs::makeDir(QDir::homeDirPath()+QString("/.nx/S-%1").arg(m_sessionID),0700);
					QFile f(QDir::homeDirPath()+QString("/.nx/S-%1/options").arg(m_sessionID));
					f.open(IO_WriteOnly); //error handling
					QTextStream stream(&f);
					stream<<"cookie="<<m_sessionCookie<<",root="+QFile::encodeName(QDir::homeDirPath()+"/.nx")<<",session=SomeHost,id="<<m_sessionID<<",connect="<<m_sshData.host<<":"<<m_sessionDisplay<<endl;
					f.close(); //error handling
					kdDebug()<<"Starting proxy"<<endl;
					m_proxyProcess=new KProcess();
					(*m_proxyProcess)<<"nxproxy" << "-S" <<QString("options=%1").arg(+QFile::encodeName(QDir::homeDirPath()+QString("/.nx/S-%1").arg(m_sessionID)+QString("/options:%1").arg(m_sessionDisplay)));
					connect(m_proxyProcess,SIGNAL(processExited(KProcess*)),this,SLOT(childExited(KProcess*)));
					m_proxyProcess->start();
//					system(s.data());
					m_state=SESSIONRUNNING;
					m_ignoreAbortConnecting=true;
					if (m_connecting) {
						m_connecting->close();
						m_connecting=0;
					}
					emit proxyStarted();
					disconnect(m_sshProcess,SIGNAL(processExited(KProcess*)),this,SLOT(childExited(KProcess*)));
				}
			break;
		case SESSIONRUNNING:
			break;
		case ERROR:
			break;
			
	}
}


void KNXConnection::childExited(KProcess* proc) {
	if (proc==m_sshProcess) {
		if (m_state==INITIALIZATION) {
			m_connectionTimeout->stop();
			if (m_connecting) {
				switch (proc->exitStatus()) {
					case 1:
							m_connecting->noSSH(m_sshData.targetData());
						break;
					case 255:
							m_connecting->accessDenied(false,m_sshData.targetData());
						break;
					default:
							m_connecting->unknownError();
				}
			}
			deleteLater();
/*		} else {
			if (m_connecting) m_connecting->unknownError();
			deleteLater();*/
		}
	} else {
		kdDebug()<<"Proxy terminated"<<endl<<"killing connection"<<endl;
//		sendString("Exit\n");
		deleteLater();
	}
}

void KNXConnection::abortConnecting() {
	m_connecting=0;
	if (m_ignoreAbortConnecting) return;
	deleteLater();
}

bool KNXConnection::waitsForInput(int type) {
	QCString s;
	s.sprintf("NX> %d ",type);
	int pos=m_data.find(s);
	if (pos!=-1) {
		m_data.remove(0,pos+/*1+strlen(NX> 105 */9);
		return true;
	}
	return false;
}

bool KNXConnection::findLineStartingWith(QCString str) {
	if (m_data.data()==0) return false;
	if (str.length()>(m_data.length()+1) ) return false;
	if (qstrncmp(m_data.data(),str.data(),str.length())==0) return true;
	if (str.length()>(m_data.length()+2) ) return false;	
	str="\n"+str;
	return (m_data.find(str)!=-1);
}

void KNXConnection::sendString(const char *str) {
	if (!m_sshProcess) {
		kdDebug()<<"there is no ssh process to send data to"<<endl;
		return;
	}
	char *data=strdup(str);
	//kdDebug()<<"ENQUEUEING TO SEND: "<<data<<endl;
	if (send0==0) {
		send0=data;
		m_sshProcess->writeStdin(data,strlen(data));
	} else {
		assert(send1==0);
		send1=data;
	}
	
}

bool KNXConnection::gotResponse(const char *str) {
	int posOfNX;
	posOfNX=m_data.find("NX> ");
	if (posOfNX<0) return false;
	posOfNX+=4;
	return (posOfNX==m_data.find(str));
}

void KNXConnection::dataHasBeenWrittenToSSH() {
	if (send0) free(send0);
	send0=send1;
	send1=0;
	if (send0) m_sshProcess->writeStdin(send0,strlen(send1));
	
}

bool KNXConnection::getValue(int id,QCString &val) {
	if (m_data.length()==0) return false;
	char searchid[50];
	sprintf(searchid,"NX> %d",id);
	int pos=m_data.find(searchid);
	if (pos<0) return false;
	pos=m_data.find(":",pos);
	if (pos<0) return false;
	int endpos=m_data.find("\n",pos);
	if (endpos<1) return false;
	pos+=2;
	if (m_data[endpos-1]=='\r') endpos--;
	if (pos>=endpos) return false;
	val=m_data.mid(pos,endpos-pos);
	kdDebug()<<"READ property: "<<id<<" with value="<<val<<"<<"<<endl;
	return true;
}

void KNXConnection::timeoutWhileConnecting() {
	m_connectionTimeout->stop();
	delete m_sshProcess;
	m_sshProcess=0;
	if (m_connecting) m_connecting->timeout();
}


void KNXConnection::load(const QString &connectionFile) {
	delete m_tunneledConnection;
	m_tunneledConnection=0;

	m_sessionString="";
	m_sshData.user="nx";
	QFile f(QFile::encodeName(connectionFile));
	f.open(IO_ReadOnly);
	QDomDocument doc;
	bool ok=doc.setContent(&f);
	f.close();
	if (!ok) return; //error handling / security here

	QDomElement docRoot=doc.documentElement();
	//Determin the Type
	m_sessionType=UnknownSession;
	for(QDomNode group=docRoot.firstChild(); !group.isNull(); group=group.nextSibling()) {
		QDomElement groupEl=group.toElement();
		if ( (groupEl.isNull()) || (groupEl.tagName()!="group")) continue;
		if ( groupEl.attribute("name")!="General") continue;
			for (QDomNode prop=groupEl.firstChild(); !prop.isNull(); prop=prop.nextSibling()) {
				QDomElement propEl=prop.toElement();
				if ( propEl.isNull() || propEl.tagName()!="option") continue;
				if (propEl.attribute("key")=="Session") {
					QString st=propEl.attribute("value");
					if (st=="Unix") m_sessionType=UnixSession;
					else if (st=="VNC") m_sessionType=VNCSession;
					else if (st=="Windows") m_sessionType=WindowsSession;
					break;
				}
			}
	}


       for(QDomNode group=docRoot.firstChild(); !group.isNull(); group=group.nextSibling()) {
                QDomElement groupEl=group.toElement();
                if ( (groupEl.isNull()) || (groupEl.tagName()!="group")) continue;
		QString groupName=groupEl.attribute("name");
		if (groupName=="Advanced") readAdvanced(groupEl);
		else if (groupName=="General") readGeneral(groupEl);
		else if ( (groupName=="VNC Session") && (m_sessionType==VNCSession)) m_tunneledConnection=new TunneledVNCConnection(groupEl);
		else if ( (groupName=="Windows Session") && (m_sessionType==WindowsSession)) m_tunneledConnection=new TunneledWindowsConnection(groupEl);
	}
}

void KNXConnection::readAdvanced(QDomElement &groupEl) {
	for (QDomNode prop=groupEl.firstChild(); !prop.isNull(); prop=prop.nextSibling()) {
		QDomElement propEl=prop.toElement();
		if ( (propEl.isNull() || propEl.tagName()!="option")) continue;
		QString key=propEl.attribute("key");
		if (key=="Cache size") m_sessionString+=" --cache=\""+propEl.attribute("value")+"M\"";
		else if (key=="Cache size on disk") m_sessionString+=" --images=\""+propEl.attribute("value")+"M\"";
		else if (key=="Enable SSL Encryption") m_sessionString+=" --encryption=\""+ QString((propEl.attribute("value")=="false")?"0":"1")+"\"";
#warning support for the rest needed
	}
}


void KNXConnection::readGeneral(QDomElement &groupEl) {
	for (QDomNode prop=groupEl.firstChild(); !prop.isNull(); prop=prop.nextSibling()) {
		QDomElement propEl=prop.toElement();
		if ( (propEl.isNull() || propEl.tagName()!="option")) continue;
		QString key=propEl.attribute("key");
		kdDebug()<<"readGeneral: key="<<key<<endl;
		if (key=="Backingstore") m_sessionString+=" --backingstore=\""+propEl.attribute("value")+"\"";
		else if (key=="Link speed") m_sessionString+=" --link=\""+propEl.attribute("value")+"\"";
		else if (key=="Server host") {m_sshData.host=propEl.attribute("value"); kdDebug()<<"have read ServerHost:"<<m_sshData.host<<endl;}
		else if (key=="Server port") m_sshData.port=propEl.attribute("value");
		else if (key=="Resolution") { 
			QString resolution=propEl.attribute("value");
			if (resolution=="fullscreen") 
				m_sessionString+=" --geometry=\""+resolution+"\"";
			else 
				m_sessionString+=" --geometry=\""+resolution+"+0+0\"";

		}
		else if (key=="Desktop") {
			if (m_sessionType==UnixSession) m_sessionString+=" --type=\"unix-"+propEl.attribute("value").lower()+"\"";
		} 		
#warning support for the rest needed

	}

	if (m_sessionType==VNCSession) {
		m_sessionString+=" --type=\"vnc\"";
	} else if (m_sessionType==WindowsSession) {
		m_sessionString+=" --type=\"windows\"";
	}
}











static void KNXdecodePassword(QString pwd,QCString &result) {
	if (!pwd.isEmpty()) {
		QStringList cl=QStringList::split(':',pwd);
			int i=1;
			for (QStringList::Iterator ci=cl.begin();ci!=cl.end();i++,++ci) {
				XKeyEvent e;
				e.display=qt_xdisplay();
				e.keycode=XKeysymToKeycode(qt_xdisplay(),((*ci).toLong()-i));	
				char decoded[21];
				KeySym dummy;
				XComposeStatus dummy2;
				XLookupString(&e,decoded,20,&dummy,&dummy2);
				result+=decoded;
//				result+=XKeysymToString(((*ci).toLong()-i));
			}
		}

}

KNXConnection::TunneledVNCConnection::TunneledVNCConnection(QDomElement &groupEl):TunneledConnection(groupEl) {
	for (QDomNode prop=groupEl.firstChild(); !prop.isNull(); prop=prop.nextSibling()) {
		QDomElement propEl=prop.toElement();
		if ( (propEl.isNull() || propEl.tagName()!="option")) continue;
		QString key=propEl.attribute("key");
		if (key=="Display") display=propEl.attribute("value");
		else if (key=="Server") host=propEl.attribute("value");
		else if (key=="Password") KNXdecodePassword(propEl.attribute("value"),password);
		else if (key=="Remember") remember=propEl.attribute("value");
	}
}

KNXConnection::TunneledVNCConnection::~TunneledVNCConnection() {}

bool KNXConnection::TunneledVNCConnection::askCredentials(QWidget* w,const QString &sessionname){
	if (password.isEmpty()) {
		do {
			if (KPasswordDialog::getPassword(password,i18n("<qt>VNC-Server password for<br><b>%1</b></qt>").arg(sessionname))!=QDialog::Accepted)
				return false;
			if (password.isEmpty()) {
				KMessageBox::sorry(w,i18n("The VNC Server password has to consist of at least one character"));
			}
		} while (password.isEmpty());
	}
	return true;
}

QString KNXConnection::TunneledVNCConnection::settingsString() {
	return QString(" --agent_password=\"%1\"").arg(password)+
		QString(" --agent_server=\"%1\"").arg(host+display);

}




KNXConnection::TunneledWindowsConnection::TunneledWindowsConnection(QDomElement &groupEl):TunneledConnection(groupEl) {
	for (QDomNode prop=groupEl.firstChild(); !prop.isNull(); prop=prop.nextSibling()) {
		QDomElement propEl=prop.toElement();
		if ( (propEl.isNull() || propEl.tagName()!="option")) continue;
		QString key=propEl.attribute("key");
		if (key=="Server") host=propEl.attribute("value");
		else if (key=="Password") KNXdecodePassword(propEl.attribute("value"),password);
		else if (key=="Remember") remember=propEl.attribute("value");
		else if (key=="User") user=propEl.attribute("value");
	}
}

KNXConnection::TunneledWindowsConnection::~TunneledWindowsConnection() {}

bool KNXConnection::TunneledWindowsConnection::askCredentials(QWidget*,const QString &sessionname){
	return WinUserPwd::getData(sessionname,user,password);
}

QString KNXConnection::TunneledWindowsConnection::settingsString() {
	return QString(" --agent_password=\"%1\"").arg(password)+
		QString(" --agent_server=\"%1\"").arg(host) +
		QString(" --agent_user=\"%1\"").arg(user);

}

#include "knxconnection.moc"
